/**
 * @author Dr. Vargas-Perez
 * Fall, 2018. Kalamazoo College
 * 
 * With help from our text book and geeksforgeeks.com
 * 
 * This class creates a AVL Binary Search Tree
 *
 */

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

public class AVLTree {

	AVLNode root;

	public AVLTree() {
		root = null;
	}

	/**
	 * Inserts a node to the AVL Tree in its 
	 * corresponding place.
	 * 
	 * @param element: key of the node to be inserted
	 */
	public void addElement(int element) {
		root = add(root, element);
	}

	/**
	 * Recursively insert a given element into the tree
	 * starting from the root of the tree 
	 * @param node: the root of this tree
	 * @param element: key of the node to be inserted
	 * 
	 * @return a pointer to node (now with the new node inserted)
	 */
	public AVLNode add(AVLNode node, int element) {
		/* Hints:
		 * 
		 * First your base case: when node is null, return a new AVLNode.
		 * 
		 * Recursively insert the element to the left of node 
		 * if smaller, or to the right of node is greater or equal.
		 * 
		 * After the new node is inserted, the backtracking needs to
		 * fix each node's height (calling maxHeight), and check the 
		 * balance factor of the node.
		 * 
		 * if balance factor is:
		 * -2 for the node and -1 for leftChild of node: Right rotation
		 * -2 for the node and +1 for leftChild of node: LeftRight rotation
		 * +2 for the node and +1 for rightChild of node: Left rotation
		 * +2 for the node and -1 for rightChild of node: RightLeft rotation
		 * 
		 * return a pointer to this node.
		 */
		
		// Base case
		if (node == null) {
			return new AVLNode(element);
		}

		return node;

	}

	/**
	 * Perform a right rotation and return a pointer to the new root.
	 * 
	 * @param root: the node to be rotated around
	 * @return a pointer to the new root.
	 */
	public AVLNode rotateRight(AVLNode root) {
		/* Hints:
		 * 
		 * Get a pointer for the left child of root: this is newRoot.
		 * Get a pointer for the right child of the newRoot: this will be the new
		 * newLeftChild of the former root.
		 * 
		 * Set root as the right child of the newRoot.
		 * Set newLeftChild as the left child of former root.
		 * 
		 * Fix the heights: first former root's height and then 
		 * newRoot's height.
		 * 
		 * return a pointer to newRoot
		 */
		
		return null;
	}

	public AVLNode rotateLeft(AVLNode root) {
		/* Hints:
		 * 
		 * Get a pointer for the right child of root: this is newRoot.
		 * Get a pointer for the left child of the newRoot: this will be the new
		 * newRightChild of the former root.
		 * 
		 * Set root as the left child of the newRoot.
		 * Set newRightChild as the right child of former root.
		 * 
		 * Fix the heights: first former root's height and then 
		 * newRoot's height.
		 * 
		 * return a pointer to newRoot
		 */
		
		return null;

	}
	/**
	 * Returns the maximum height between the left and right
	 * subtrees of the node
	 * @param node
	 * @return
	 */
	public int maxHeight(AVLNode node) {
		/* Hints:
		 * 
		 * Return 0 if node is null
		 * 
		 * Get height of left child node: if node is null height is 0, 
		 * else height is getNodeHeight + 1
		 * 
		 * Get height of right child node: if node is null height is 0, 
		 * else height is getNodeHeight + 1
		 * 
		 * Return the highest height.
		 * 
		 * Note: Because each nodeHeight is initialized to 0 when newly
		 * created, you need to add 1 to get the height for its parent.
		 * 
		 */

		return 0;
	}
	/**
	 * Returns the balance factor of this node, that is the difference
	 * in height between the right child and the left child:
	 * -int: means long path to the left
	 * +int: means long path to the right
	 *
	 * @param node
	 * @return the balance factor of this node
	 */
	public int balanceFactor(AVLNode node) {
		/* Hints:
		 * 
		 * Return 0 if node is null
		 * 
		 * Get height of left child node: if node is null height is 0, 
		 * else height is getNodeHeight + 1
		 * 
		 * Get height of right child node: if node is null height is 0, 
		 * else height is getNodeHeight + 1
		 * 
		 * Return the right child height - left child height.
		 * 
		 * Note: Because each nodeHeight is initialized to 0 when newly
		 * created, you need to add 1 to get the height for its parent.
		 * 
		 */

		return 0;
	}
	
	/**
	 * A lever-order (breadth first) traversal of this tree. It
	 * prints "*" for null nodes.
	 */
	public void levelOrderTraversal()
	{
		Queue<Object> queue = new LinkedList<Object>();
		queue.add(root);
		while( ! queue.isEmpty() )
		{
			AVLNode tree = (AVLNode)queue.remove();
			if (tree != null)
			{
				System.out.println(tree.getElement());
				queue.add(tree.getLeftChild());
				queue.add(tree.getRightChild());
			}
			else {
				System.out.println("*");
			}
		}
	}	

	public static void main(String[] args) {
		AVLTree tree = new AVLTree();
		
		// Inserts ten random integers (from 0 to 99) in the AVL tree
		System.out.print("Numbers to be inserted: ");
		for (int i = 0; i < 10; i++) {
			int x = (int)(Math.random() * 100);
			System.out.print(x + " ");
			tree.addElement(x);
		}  
		
		System.out.println();
		System.out.println("Now printing a pretty AVL: \n");
		
		PrintBinaryTree print = new PrintBinaryTree(tree.root);

		print.printTree();
	}

}
